/*
 * Copyright 2010 - org.tinyjee.maven
 *
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */

package org.tinyjee.maven.dim;

import org.apache.maven.doxia.logging.Log;
import org.apache.maven.doxia.macro.MacroExecutionException;
import org.apache.maven.doxia.sink.Sink;
import org.apache.maven.doxia.sink.SinkEventAttributeSet;

import static org.tinyjee.maven.dim.IncludeMacroSignature.*;

import org.codehaus.plexus.util.StringUtils;
import org.tinyjee.maven.dim.sh.BrushResolver;
import org.tinyjee.maven.dim.sh.HighlighterThreadLocal;
import org.tinyjee.maven.dim.sh.StyleHandler;
import org.tinyjee.maven.dim.sources.ContentCodec;
import org.tinyjee.maven.dim.sources.Source;
import org.tinyjee.maven.dim.utils.MacroLoggers;
import org.tinyjee.maven.dim.utils.XhtmlEntityResolver;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import java.io.File;
import java.io.IOException;
import java.io.StringReader;
import java.io.StringWriter;
import java.net.URL;
import java.util.*;

/**
 * Performs code highlighting on the provided source before writing it to the Sink.
 *
 * @author Juergen_Kellerer, 2010-09-04
 * @version 1.0
 */
public class CodeHighlighter {

	static final String LB = System.getProperty("line.separator", "\n");

	static final Map<Sink, Boolean> styleWasAdded = Collections.synchronizedMap(new WeakHashMap<Sink, Boolean>());
	static final HighlighterThreadLocal highlighters = new HighlighterThreadLocal();

	StyleHandler styleHandler = new StyleHandler();

	static final String HTML_OPEN_TAG =
			"<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\" " + LB +
					"\"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd\"> " + LB +
					"<html xmlns=\"http://www.w3.org/1999/xhtml\" xml:lang=\"en\"> " + LB;
	static final String HTML_CLOSE_TAG = "</html>";

	final DocumentBuilderFactory documentBuilderFactory = DocumentBuilderFactory.newInstance();
	final TransformerFactory transformerFactory = TransformerFactory.newInstance();


	/**
	 * Processes the given source lines and writes the results into the sink.
	 *
	 * @param source     The source used to obtain the content.
	 * @param targetPath the path to process the content to (may be unknown).
	 * @param sink       the target sink to write to.
	 * @throws MacroExecutionException in case of the process fails.
	 */
	public void execute(Source source, File targetPath, Sink sink)
			throws MacroExecutionException {
		try {
			Source.Content content = source.getContent();
			execute(source.getURL(), content.getParameters(), content.getSections(), targetPath, sink);
		} catch (IOException e) {
			throw new MacroExecutionException(e.getMessage(), e);
		}
	}

	/**
	 * Processes the given source lines and writes the results into the sink.
	 *
	 * @param source            The source URL used to obtain the content.
	 * @param sourceLines       a map of first line number to lines mapping.
	 * @param requestParameters the original request params.
	 * @param targetPath        the path to process the content to (may be unknown).
	 * @param sink              the target sink to write to.
	 * @throws MacroExecutionException in case of the process fails.
	 */
	public void execute(URL source, Map requestParameters,
	                    Map<Integer, List<String>> sourceLines,
	                    File targetPath, Sink sink) throws MacroExecutionException {

		final Map<String, String> params = buildHighlighterParams(source, sourceLines, requestParameters);
		final String brushName = findBrushName(source, sourceLines, requestParameters);

		final List<String> lines = new LinkedList<String>();
		final List<Integer> linesToRemove = new LinkedList<Integer>();
		final int firstLine = requestParameters.containsKey(PARAM_SET_FIRST_LINE) ?
				Integer.parseInt((String) requestParameters.get(PARAM_SET_FIRST_LINE)) : -1;

		int lineNumber = -1, delta = 0;
		for (Map.Entry<Integer, List<String>> entry : sourceLines.entrySet()) {
			if (lines.isEmpty()) {
				delta = firstLine == -1 ? 0 : entry.getKey() - firstLine;
				params.put("first-line", String.valueOf(entry.getKey() - delta));
			} else {
				for (int i = lineNumber, len = entry.getKey(); i < len; i++) {
					linesToRemove.add(i - delta);
					lines.add("");
				}
			}
			lineNumber = entry.getKey() + entry.getValue().size();
			lines.addAll(entry.getValue());
		}

		final String code = ContentCodec.write(lines);
		if (code.length() == 0) {
			MacroLoggers.getInstance().get().error(
					"No content to highlight for source '" + source + "' with params '" + params + '\'');
			return;
		}

		String htmlResult = highlighters.getHighlighter().highlightToHTML(brushName, code, params);
		if (!linesToRemove.isEmpty())
			htmlResult = removeLines(htmlResult, linesToRemove);

		final Object highlightTheme = requestParameters.get(PARAM_HIGHLIGHT_THEME);

		if (isHTMLCapableSink(sink)) {
			addStyleIfNeeded(sink, (highlightTheme == null ? null : highlightTheme.toString()), targetPath);
			sink.rawText(htmlResult);
		} else {
			// TODO: Render image instead (e.g. using Lobo or Swing).
			sink.verbatim(SinkEventAttributeSet.BOXED);
			sink.text("TODO: Image Rendering !!");
			sink.verbatim_();
		}
	}

	/**
	 * Finds the name of the code highlighter to use.
	 *
	 * @param source            the source URL.
	 * @param sourceLines       the source content.
	 * @param requestParameters the request parameters of the macro request.
	 * @return the name of the code highlighter to use or 'plain' if not found.
	 */
	protected String findBrushName(URL source, Map<Integer, List<String>> sourceLines, Map requestParameters) {
		String brushName = findBrushAlias(source, sourceLines, requestParameters);
		int idx = brushName.indexOf('/');
		return idx == -1 ? brushName : brushName.substring(idx + 1);
	}

	/**
	 * Finds the name of the code highlighter to use.
	 *
	 * @param source            the source URL.
	 * @param sourceLines       the source content.
	 * @param requestParameters the request parameters of the macro request.
	 * @return the name of the code highlighter to use or 'plain' if not found.
	 */
	protected String findBrushAlias(URL source, Map<Integer, List<String>> sourceLines, Map requestParameters) {
		String brushAlias = requestParameters == null ? null :
				(String) requestParameters.get(PARAM_HIGHLIGHT_TYPE);

		if (brushAlias == null)
			for (BrushResolver resolver : highlighters.getHighlighter().getBrushResolvers()) {
				brushAlias = resolver.detectBrushAlias(source, sourceLines);
				if (brushAlias != null)
					break;
			}

		return brushAlias == null ? "plain" : brushAlias;
	}

	/**
	 * Returns true if the source content is HTML with embedded source code of another language.
	 *
	 * @param source            the source URL.
	 * @param sourceLines       the source content.
	 * @param requestParameters the request parameters of the macro request.
	 * @return true if the source content is HTML with embedded source code of another language.
	 */
	protected boolean isCodeSurroundedByHTML(URL source, Map<Integer, List<String>> sourceLines,
	                                         Map requestParameters) {
		return findBrushAlias(source, sourceLines, requestParameters).startsWith("html/");
	}

	/**
	 * Builds a map of params that can be passed to the SyntaxHighlighter library.
	 *
	 * @param source            the source URL.
	 * @param sourceLines       the source content.
	 * @param requestParameters the request parameters of the macro request.
	 * @return a map of params that can be passed to the SyntaxHighlighter library.
	 */
	protected Map<String, String> buildHighlighterParams(URL source, Map<Integer, List<String>> sourceLines,
	                                                     Map requestParameters) {
		final Map<String, String> params = new HashMap<String, String>();

		// Decide whether the code is embedded into HTML or XHTML.
		params.put("html-script", String.valueOf(isCodeSurroundedByHTML(source, sourceLines, requestParameters)));

		// Disable the gutter if explicitly requested.
		if ("false".equalsIgnoreCase(String.valueOf(requestParameters.get(PARAM_SHOW_GUTTER)).trim()))
			params.put("gutter", "false");

		// Pass through the setting of padded line numbers.
		Object padLineNumbers = requestParameters.get(PARAM_PAD_LINE_NUMBERS);
		if (padLineNumbers != null)
			params.put("pad-line-numbers", Integer.valueOf(padLineNumbers.toString()).toString());

		// Parse highlighted lines
		Object highlightLines = requestParameters.get(PARAM_HIGHLIGHT);
		if (highlightLines != null) {
			StringBuilder buffer = new StringBuilder().append("[0");
			String[] lines = highlightLines.toString().split("\\s*[,;+]\\s*");
			for (String lineNumber : lines) {
				String[] parts = lineNumber.split("\\s*-\\s*");
				if (parts.length == 1)
					buffer.append(", ").append(Integer.parseInt(parts[0]));
				else if (parts.length == 2 && StringUtils.isNotEmpty(parts[0])) {
					for (int i = Integer.parseInt(parts[0]), len = Integer.parseInt(parts[1]); i <= len; i++)
						buffer.append(", ").append(i);
				}
			}
			params.put("highlight", buffer.append(']').toString());
		}

		// Pass through any params that were prefixed with "sh-" (removing the prefix).
		for (Object entry : requestParameters.entrySet()) {
			Map.Entry e = (Map.Entry) entry;
			String key = String.valueOf(e.getKey());
			if (key.startsWith("sh-"))
				params.put(key.substring(3), (String) e.getValue());
		}

		return params;
	}

	/**
	 * Returns true if the sink produces some sort of HTML.
	 *
	 * @param sink the sink to test.
	 * @return true if the sink produces some sort of HTML.
	 */
	public boolean isHTMLCapableSink(Sink sink) {
		try {
			return sink instanceof org.apache.maven.doxia.markup.HtmlMarkup;
		} catch (NoClassDefFoundError ignored) {
			// Old API, use String-Search as fallback.
			String[] names = {"XhtmlSink", "XhtmlBaseSink"};
			for (Class<?> cls = sink.getClass(); cls != null; cls = cls.getSuperclass())
				for (String name : names)
					if (cls.getName().contains(name))
						return true;
			return false;
		}
	}

	/**
	 * Handles the addition of CSS styles.
	 *
	 * @param sink         the target sink to add the styles to.
	 * @param theme        the theme to use.
	 * @param targetFolder the folder that may be used to copy the CSS files inside.
	 * @throws MacroExecutionException in case of the operation is executed on a NON-HTML sink.
	 */
	protected void addStyleIfNeeded(Sink sink, String theme, File targetFolder) throws MacroExecutionException {
		if (!isHTMLCapableSink(sink)) {
			throw new MacroExecutionException("Cannot add CSS styles to sink '" +
					sink.getClass() + "', it's not capable of it.");
		}

		if (!Boolean.TRUE.equals(styleWasAdded.get(sink))) {

			if (theme != null && !styleHandler.isHandlingStyleForTheme(theme))
				styleHandler = new StyleHandler(theme);

			boolean inlineCSS = targetFolder == null;

			sink.rawText(styleHandler.getHTML(inlineCSS));
			if (!inlineCSS)
				styleHandler.copySourcesIfMissing(targetFolder);

			styleWasAdded.put(sink, Boolean.TRUE);
		}
	}

	protected String removeLines(String content, Collection<Integer> linesToRemove) {
		try {
			final Set<String> classesToRemove = new HashSet<String>(linesToRemove.size());
			for (Integer integer : linesToRemove)
				classesToRemove.add("number" + integer);

			content = HTML_OPEN_TAG + content + HTML_CLOSE_TAG;
			DocumentBuilder builder = documentBuilderFactory.newDocumentBuilder();
			builder.setEntityResolver(new XhtmlEntityResolver());
			Document document = builder.parse(new InputSource(new StringReader(content)));
			NodeList divs = document.getDocumentElement().getElementsByTagName("div");
			for (int i = divs.getLength() - 1; i >= 0; i--) {
				Element div = (Element) divs.item(i);
				for (String className : div.getAttribute("class").split("\\s+")) {
					if (classesToRemove.contains(className))
						div.getParentNode().removeChild(div);
				}
			}

			StringWriter buffer = new StringWriter(content.length());
			transformerFactory.newTransformer().transform(new DOMSource(document), new StreamResult(buffer));
			content = buffer.toString();
			content = content.substring(content.indexOf('<', content.indexOf("<html") + 5),
					content.lastIndexOf(HTML_CLOSE_TAG));
		} catch (Exception e) {
			Log log = MacroLoggers.getInstance().get();
			log.error("Failed to remove hidden lines from highlighted content. " +
					"The result may contain content that should have been hidden.", e);
			log.error(content);
		}
		return content;
	}
}
