/*
 * Copyright 2010 - org.tinyjee.maven
 *
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */

package org.tinyjee.maven.dim;

import org.apache.maven.doxia.logging.Log;
import org.apache.maven.doxia.logging.SystemStreamLog;
import org.apache.maven.doxia.macro.AbstractMacro;
import org.apache.maven.doxia.macro.MacroExecutionException;
import org.apache.maven.doxia.macro.MacroRequest;
import org.apache.maven.doxia.module.apt.AptParser;
import org.apache.maven.doxia.module.fml.FmlParser;
import org.apache.maven.doxia.module.xdoc.XdocParser;
import org.apache.maven.doxia.parser.ParseException;
import org.apache.maven.doxia.parser.Parser;
import org.apache.maven.doxia.sink.Sink;
import org.apache.maven.doxia.sink.SinkEventAttributeSet;
import org.codehaus.plexus.PlexusConstants;
import org.codehaus.plexus.PlexusContainer;
import org.codehaus.plexus.component.repository.exception.ComponentLifecycleException;
import org.codehaus.plexus.component.repository.exception.ComponentLookupException;
import org.codehaus.plexus.context.Context;
import org.codehaus.plexus.context.ContextException;
import org.codehaus.plexus.personality.plexus.lifecycle.phase.Contextualizable;
import org.codehaus.plexus.util.FileUtils;
import org.codehaus.plexus.util.StringUtils;
import org.tinyjee.maven.dim.sources.*;
import org.tinyjee.maven.dim.utils.MacroLoggers;
import org.tinyjee.maven.dim.utils.ResolveUtils;

import java.io.File;
import java.io.StringReader;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.HashMap;
import java.util.Map;

import static org.tinyjee.maven.dim.IncludeMacroSignature.*;
import static org.tinyjee.maven.dim.utils.ResolveUtils.isAbsolute;

/**
 * Implements a more feature rich alternative to the bundled "Snippet" macro.
 *
 * @author Juergen_Kellerer, 2010-09-03
 * @version 1.0
 * @plexus.component role="org.apache.maven.doxia.macro.Macro" role-hint="include"
 */
public class IncludeMacro extends AbstractMacro implements Contextualizable {

	/**
	 * Specifies the folder name of the folder where resources are put.
	 */
	public static final String RESOURCES_FOLDER_NAME = "resources";

	/**
	 * Specifies the extension used to detect velocity templates.
	 */
	public static final String VELOCITY_TEMPLATE_EXTENSION = ".vm";

	final CodeHighlighter highlighter = new CodeHighlighter();
	private PlexusContainer container;

	/**
	 * {@inheritDoc}
	 */
	public void contextualize(Context context) throws ContextException {
		container = (PlexusContainer) context.get(PlexusConstants.PLEXUS_KEY);
	}

	/**
	 * {@inheritDoc}
	 */
	@SuppressWarnings("unchecked")
	public void execute(Sink sink, MacroRequest request) throws MacroExecutionException {
		MacroLoggers.getInstance().set(getLogger());

		final File basePath = request.getBasedir();
		final Map parameters = request.getParameters();

		for (RequestParameterTransformer transformer : RequestParameterTransformer.TRANSFORMERS)
			transformer.transformParameters(parameters);
		final Map<String, String> textParameters = extractTextParameters(parameters);

		URL sourceURL = null;

		String sourceClass = textParameters.get(PARAM_SOURCE_CLASS);
		String rawSource = textParameters.get(PARAM_SOURCE);
		if (!StringUtils.isEmpty(rawSource))
			sourceURL = ResolveUtils.findSource(basePath, rawSource);

		String file = textParameters.get(PARAM_FILE);
		if (!StringUtils.isEmpty(file)) {
			try {
				sourceURL = (isAbsolute(file) ? new File(file) : new File(basePath, file)).toURI().toURL();
			} catch (MalformedURLException e) {
				throw new MacroExecutionException(file, e);
			}
		}

		Source source = sourceClass == null ? null :
			new SourceClassAdapter(basePath, parameters, sourceClass);

		if (sourceURL != null) {
			Source urlSource = new URLSource(sourceURL, (Map) textParameters);
			source = source == null ? urlSource : new MergeSource(urlSource, source);

			String rawSourceIsTemplate = textParameters.get(PARAM_SOURCE_IS_TEMPLATE);
			if (sourceURL.getFile().toLowerCase().endsWith(VELOCITY_TEMPLATE_EXTENSION) ||
				Boolean.parseBoolean(rawSourceIsTemplate)) {
				// Removing the ".VM" from the source to allow get right results in extension based type detection.
				try {
					sourceURL = new URL(sourceURL.toExternalForm().replace(VELOCITY_TEMPLATE_EXTENSION, ""));
				} catch (MalformedURLException e) {
					throw new RuntimeException(e);
				}
				// Allow to force disabling template processing in order to include templates themselves.
				boolean templateIsDisabled = rawSourceIsTemplate != null && !Boolean.parseBoolean(rawSourceIsTemplate);
				if (!templateIsDisabled)
					source = new TemplateSource(sourceURL, source);
			}
		}

		if (source == null) {
			throw new IllegalArgumentException("One of the parameters 'source', 'file' or " +
				"'source-class' is required.");
		}

		// Import content params
		for (Map.Entry<String, Object> entry : source.getContent().getParameters().entrySet()) {
			if (entry.getValue() != null)
				textParameters.put(entry.getKey(), entry.getValue().toString());
		}

		processSource(textParameters, basePath, source, sink);
	}

	/**
	 * Processes the given source to the Sink.
	 *
	 * @param textParameters a text-only version of the request params.
	 * @param basePath	   the module's base path.
	 * @param source		 the effective source.
	 * @param sink		   the sink to write the output to.
	 * @throws MacroExecutionException in case of the processing failed.
	 */
	protected void processSource(Map<String, String> textParameters, File basePath,
								 Source source, Sink sink) throws MacroExecutionException {
		Log log = MacroLoggers.getInstance().get();
		Source.Content content = source.getContent();
		Parser parser = getContentParser(source.getURL(), textParameters);

		Boolean verbatim = textParameters.containsKey(PARAM_VERBATIM) ?
			Boolean.valueOf(textParameters.get(PARAM_VERBATIM)) :
			parser == null;

		if (verbatim) {
			File cssTargetPath = null, sitePath = findSiteSourceDirectory(basePath, textParameters);

			if (sitePath != null) {
				cssTargetPath = new File(sitePath, RESOURCES_FOLDER_NAME);
				if (!cssTargetPath.isDirectory() && !cssTargetPath.mkdirs()) {
					log.warn("Failed creating resource path '" + cssTargetPath + "', will inline CSS styles instead.");
					cssTargetPath = null;
				}
			}

			try {
				if ("none".equalsIgnoreCase(textParameters.get(PARAM_HIGHLIGHT_TYPE))) {
					sink.verbatim(SinkEventAttributeSet.BOXED);
					sink.rawText(ContentCodec.write(content.getSections()));
					sink.verbatim_();
				} else
					highlighter.execute(source, cssTargetPath, sink);
			} catch (Exception e) {
				throw new MacroExecutionException(source.getURL().toString(), e);
			}
		} else {
			String textContent = null;
			try {
				textContent = ContentCodec.write(content.getSections());
				if (parser == null)
					sink.rawText(textContent);
				else
					parser.parse(new StringReader(textContent), sink);
			} catch (ParseException e) {
				throw new MacroExecutionException(
					"Failed to parse content [\n" + textContent + "\n] loaded from " + source.getURL() +
						" with parser " + parser.getClass(), e);
			} catch (Exception e) {
				throw new MacroExecutionException(source.getURL().toString(), e);
			}
		}
	}

	/**
	 * Finds the directory where the site sources are located.
	 *
	 * @param baseDir the base directory ot the module.
	 * @param params  the parameters of the macro request.
	 * @return the directory where the site sources are located.
	 */
	protected File findSiteSourceDirectory(File baseDir, Map<String, String> params) {
		String siteDirectory = params.get(PARAM_SITE_DIRECTORY);
		File proposedSiteDir = StringUtils.isEmpty(siteDirectory) ? null : new File(baseDir, siteDirectory);
		return ResolveUtils.findSiteSourceDirectory(baseDir, proposedSiteDir);
	}

	/**
	 * Extracts all text only parameters from the given map of request parameters.
	 *
	 * @param requestParameters the request parameters.
	 * @return A map containing only strings.
	 */
	protected Map<String, String> extractTextParameters(Map requestParameters) {
		final Map<String, String> textParameters = new HashMap<String, String>(requestParameters.size());
		for (Object o : requestParameters.entrySet()) {
			Map.Entry entry = (Map.Entry) o;
			textParameters.put((String) entry.getKey(), String.valueOf(entry.getValue()));
		}

		return textParameters;
	}

	/**
	 * Returns a content parser for the given source or 'null' if not no parser exists.
	 *
	 * @param source the source to retrieve the parser for.
	 * @param params the request params.
	 * @return a content parser for the given source or 'null' if not no parser exists.
	 */
	protected Parser getContentParser(URL source, Map<String, String> params) {
		String type = params.get(PARAM_SOURCE_CONTENT_TYPE);
		if (type == null && source != null)
			type = FileUtils.extension(source.getFile());
		if (type != null) {
			if ("apt".equalsIgnoreCase(type))
				return new AptParser();
			else if ("xdoc".equalsIgnoreCase(type))
				return new XdocParser();
			else if ("fml".equalsIgnoreCase(type))
				return new FmlParser();
		}

		return null;
	}

	@SuppressWarnings("unchecked")
	private <E extends Parser> E lookupParser(String role, E fallbackInstance) {
		final Log log = MacroLoggers.getInstance().get();
		try {
			if (container != null)
				return (E) container.lookup("org.apache.maven.doxia.parser.Parser", role);
		} catch (ClassCastException e) {
			log.error("Parser for role '" + role + "' was of an unexpected class.", e);
		} catch (ComponentLookupException e) {
			log.error("Parser for role '" + role + "' was not found in plexus container.", e);
		}
		return fallbackInstance;
	}

	protected void releaseContentParser(Parser parser) {
		if (container != null && parser != null)
			try {
				container.release(parser);
			} catch (ComponentLifecycleException ignored) {
				MacroLoggers.getInstance().get().warn("Failed releasing parser: " + parser);
			}
	}

	/**
	 * Is a helper method to retrieve a Doxia logger using reflection
	 * (to be compatible with earlier versions of Doxia).
	 *
	 * @return a Doxia logger.
	 */
	protected Log getLogger() {
		// Using reflection as getLog() was not available in earlier versions of the API.
		try {
			return (Log) getClass().getMethod("getLog").invoke(this);
		} catch (Exception ignored) {
			return new SystemStreamLog();
		}
	}
}
