/*
 * Copyright 2010 - org.tinyjee.maven
 *
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */

package org.tinyjee.maven.dim;

/**
 * Defines the signature of the include macro.
 *
 * @author Juergen_Kellerer, 2010-09-06
 * @version 1.0
 */
public interface IncludeMacroSignature {

	/**
	 * Sets the snippet-id to use for extracting snippets out of the specified source.
	 * <p/>
	 * Snippets are processed on a line by line basis and multiple snippets with the same
	 * name may appear in one file. In order to define a section using a snippet id,
	 * surround a block with:<br/>
	 * <code>
	 * // START SNIPPET: snippet-id<br/>
	 * ... code to include ...<br/>
	 * // END SNIPPET: snippet-id<br/>
	 * </code>
	 */
	String PARAM_SNIPPET_ID = "id";

	/**
	 * Sets the source path or URL to include.
	 * <p/>
	 * Sources are searched at multiple locations including the java and site source and resource
	 * directories. In case of the specified value starts with <b>http</b> or <b>ftp</b> the source
	 * content is downloaded using a remote connection.
	 * <p/>
	 * Sources can also be loaded from the module or site building classpath using
	 * "<code>classpath:/path/to/file</code>".
	 * <br/>
	 * The classpath is constructed out of the module <code>sources</code>
	 * and <code>test-sources</code> folder without adding any dependencies defined inside the
	 * <code>pom.xml</code>. The parent classpath is defined by the dependency list of
	 * the <code>maven-site-plugin</code> and is used to lookup the source when it is
	 * not found in the module paths.
	 */
	String PARAM_SOURCE = "source";

	/**
	 * Sets a class to use as dynamic source for content, request parameters or both.
	 * <p/>
	 * A class that is set using this parameter must implement the interface <code>java.util.Map</code>
	 * (or derive from a Map implementation like <code>java.util.HashMap</code>).
	 * All entries of this map are treated as if they had been specified as request parameters,
	 * on the Macro (therefore all parameters except 'source-class' can be set dynamically).
	 * <p/>
	 * The source class may choose to use one of the 4 constructor signatures: <code>Default</code>,
	 * <code>Constructor(java.io.File basePath)</code>, <code>Constructor(java.util.Map requestProperties)</code>
	 * or <code>Constructor(java.io.File basePath, java.util.Map requestProperties)</code>.  
	 * <p/>
	 * In case of the source class wants to return the content to include (instead of loading
	 * the template specified with <code>"source"</code>), it must either return a
	 * <code>String</code> or <code>Map&lt;Integer, List&lt;String&gt;&gt;</code> when it is
	 * queried for the <code>"content"</code> key.
	 * <p/>
	 * The classpath that is used to load the class is similar to the classpath used when
	 * loading sources. (see <code>"source"</code> parameter for more information)
	 */
	String PARAM_SOURCE_CLASS = "source-class";

	/**
	 * Sets the source path using <code>java.io.File</code> instead of evaluating it against URL
	 * or File paths.
	 * <p/>
	 * Under normal conditions it's easier to set the <code>"source"</code> parameter instead.
	 * However there may be occasions where it's necessary to force that the path is treated
	 * as a file.
	 */
	String PARAM_FILE = "file";

	/**
	 * Toggles whether the source content is treated as a Velocity template.
	 * (<i>Default: auto (= true if file extension ends with ".vm")</i>)
	 */
	String PARAM_SOURCE_IS_TEMPLATE = "source-is-template";

	/**
	 * Sets the content type of the source.
	 * (<i>Default: auto (= detected by file extensions ".apt", ".fml", ".xdoc")</i>)
	 * <p/>
	 * Setting this property has no effect for verbatim content. With non-verbatim
	 * content this information is used to select a doxia parser that transcodes the
	 * the content when including it into the site. Using a parser allows to use <i>xdoc</i>
	 * templates inside <i>apt</i> documents and vice versa.
	 * <p/>
	 * Values that will select a parser are "<b>apt</b>", "<b>fml</b>", "<b>xdoc</b>",
	 * any other values will have no effect.
	 */
	String PARAM_SOURCE_CONTENT_TYPE = "source-content-type";

	/**
	 * Defines the charset to use when loading the source content.
	 * (<i>Default: UTF-8</i>)
	 */
	String PARAM_CHARSET = "charset";

	/**
	 * Toggles whether the content is verbatim (included as source) or non-verbatim (included
	 * as markup and interpreted).
	 * (<i>Default: auto</i>)
	 */
	String PARAM_VERBATIM = "verbatim";

	/**
	 * Toggles whether the gutter including line numbers is shown for verbatim content.
	 * (<i>Default: true</i>)
	 */
	String PARAM_SHOW_GUTTER = "show-gutter";

	/**
	 * Sets the line number to use for the first line (if gutter is shown).
	 * (<i>Default: auto</i>)
	 */
	String PARAM_SET_FIRST_LINE = "set-first-line";

	/**
	 * Pads all line numbers by the specified amount of zeros (if gutter is shown).
	 */
	String PARAM_PAD_LINE_NUMBERS = "pad-line-numbers";

	/**
	 * Sets a list of lines to highlight inside the verbatim content.
	 * <p/>
	 * Lines can be specified as a comma separated list of line numbers, as a range
	 * using x-y or as a mixture of the 2.<br/>
	 * Example usage: <code>highlight=1,5,10,20-25</code>
	 */
	String PARAM_HIGHLIGHT = "highlight";

	/**
	 * Sets the highlighter (brush) to use when highlighting verbatim content.
	 * (<i>Default: auto (uses source content or extension to detect)</i>)
	 * <p/>
	 * In addition to the brushes bundled with the syntax highlighting library "SyntaxHighlighter"
	 * there are 2 more brushes bundled with this macro <b>bat</b> and <b>properties</b>.
	 * <p/>
	 * Highlighting of verbatim content can be disabled by setting this property to <b>none</b>.
	 */
	String PARAM_HIGHLIGHT_TYPE = "highlight-type";

	/**
	 * Sets the highlighting color theme to use.
	 * (<i>Default: default</i>)
	 * <p/>
	 * Possible values: "<code>default</code>", "<code>django</code>", "<code>eclipse</code>",
	 * "<code>emacs</code>", "<code>fadeToGrey</code>", "<code>MDUltra</code>",
	 * "<code>midnight</code>", "<code>RDark</code>".
	 * (See: <a href="http://alexgorbatchev.com/SyntaxHighlighter/manual/themes/">
	 * http://alexgorbatchev.com/SyntaxHighlighter/manual/themes/</a>)
	 * <p/>
	 * The theme can be set only once per page. The first highlighted inclusion will
	 * set the theme for the page.
	 */
	String PARAM_HIGHLIGHT_THEME = "highlight-theme";

	/**
	 * Forces the re-loading of source content when set to <code>true</code>.
	 * <p/>
	 * Content is cached for the lifetime of the build to avoid double
	 * downloads of remote resources.
	 */
	String PARAM_NO_CACHE = "no-cache";

	/**
	 * Sets the path of the site source directory.
	 * <p/>
	 * Setting this property is not required if the site directory uses standard paths
	 * or is located directly below the module's base directory.
	 */
	String PARAM_SITE_DIRECTORY = "site-directory";
}
