/*
 * Copyright 2010 - org.tinyjee.maven
 *
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */

package org.tinyjee.maven.dim.sources;

import org.codehaus.plexus.util.IOUtil;
import org.codehaus.plexus.util.StringUtils;

import java.io.IOException;
import java.io.LineNumberReader;
import java.io.Reader;
import java.io.StringReader;
import java.util.*;

/**
 * Is a fork of "org.apache.maven.doxia.macro.snippet.SnippetReader" supporting multiple snippets of the
 * same id and different indentation chars.
 *
 * @author Juergen_Kellerer, 2010-09-06
 * @version 1.0
 */
public class ContentCodec {

	/**
	 * Defines the system dependent line break to use.
	 */
	public static final String LB = System.getProperty("line.separator", "\n");

	/**
	 * Defines the indentation chars to look after.
	 */
	private static final String INDENT_CHARS = System.getProperty("org.tinyjee.maven.dim.indentChars", "\t ");

	/**
	 * Defines the increment width that should be measured for the relevant char.
	 */
	private static final int[] INDENT_WIDTH;

	static {
		String[] widths = System.getProperty("org.tinyjee.maven.dim.indentWidths", "4, 1").split("\\s*,\\s*");
		INDENT_WIDTH = new int[widths.length];
		for (int i = 0; i < widths.length; i++)
			INDENT_WIDTH[i] = Integer.parseInt(widths[i]);
	}


	/**
	 * Reads the snippets from the given reader.
	 *
	 * @param snippetId The id of the snippet or 'null' if snippet handling shall be disabled.
	 * @param content   the raw content.
	 * @return the decoded content.
	 */
	public static Map<Integer, List<String>> read(String snippetId, String content) {
		try {
			return read(snippetId, new StringReader(content));
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
	}

	/**
	 * Reads the snippets from the given reader.
	 *
	 * @param snippetId The id of the snippet or 'null' if snippet handling shall be disabled.
	 * @param reader	the reader containing the raw content.
	 * @return the decoded content.
	 * @throws IOException in case of the read operation failed.
	 */
	public static Map<Integer, List<String>> read(String snippetId, Reader reader) throws IOException {
		ContentCodec codec = new ContentCodec(snippetId, reader);
		return codec.readSource();
	}

	/**
	 * Flattens the given content to pure text using system specific line breaks.
	 *
	 * @param content the content to write.
	 * @return a String of containing all content.
	 */
	public static String write(Map<Integer, List<String>> content) {
		StringBuilder out = new StringBuilder(255);
		for (Map.Entry<Integer, List<String>> entry : content.entrySet()) {
			if (out.length() > 0)
				out.append(LB);
			write(entry.getValue(), out);
		}
		return out.toString();
	}

	/**
	 * Flattens the given lines to pure text using system specific line breaks.
	 *
	 * @param lines the lines to write.
	 * @return a String of containing all content.
	 */
	public static String write(List<String> lines) {
		return write(lines, new StringBuilder(100 * lines.size())).toString();
	}

	/**
	 * Flattens the given lines to pure text using system specific line breaks.
	 *
	 * @param lines the lines to write.
	 * @param out   The StringBuilder instance to write to.
	 * @return The StringBuilder instance.
	 */
	public static StringBuilder write(List<String> lines, StringBuilder out) {
		for (String line : lines) {
			if (out.length() > 0)
				out.append(LB);
			out.append(line);
		}
		return out;
	}

	final Reader reader;
	final String snippetId;

	/**
	 * Reads the source out of the
	 *
	 * @param snippetId The id of the snippet or 'null' if snippet handling shall be disabled.
	 * @param reader	the reader containing the raw content.
	 */
	protected ContentCodec(String snippetId, Reader reader) {
		this.snippetId = snippetId;
		this.reader = reader;
	}

	/**
	 * Reads the snippet with given id.
	 *
	 * @return The snippet.
	 * @throws java.io.IOException if something goes wrong.
	 */
	protected Map<Integer, List<String>> readSource() throws IOException {

		Map<Integer, List<String>> allLines = readLines(snippetId);

		// Un-Indent snippets by lowest common indent
		int minIndent = Integer.MAX_VALUE;
		for (List<String> lines : allLines.values()) {
			int in = minIndent(lines);
			if (minIndent > in)
				minIndent = in;
		}

		for (List<String> lines : allLines.values())
			for (ListIterator<String> iterator = lines.listIterator(); iterator.hasNext();)
				iterator.set(unIndent(iterator.next(), minIndent));

		return allLines;
	}

	/**
	 * Returns the minimal indent of all the lines in the given List.
	 *
	 * @param lines A List of lines.
	 * @return the minimal indent.
	 */
	int minIndent(List<String> lines) {
		int minIndent = Integer.MAX_VALUE;
		for (String line : lines)
			minIndent = Math.min(minIndent, detectIndentWith(line));
		return minIndent;
	}

	/**
	 * Returns the indent of the given line.
	 *
	 * @param line A line.
	 * @return the indent.
	 */
	int detectIndentWith(String line) {
		for (int indent = 0, i = 0, len = line.length(); i < len; i++) {
			int idx = INDENT_CHARS.indexOf(line.charAt(i));
			if (idx == -1)
				return indent;
			indent += INDENT_WIDTH[idx];
		}
		return Integer.MAX_VALUE;
	}

	String unIndent(String line, int indentWidth) {
		int indent = 0;
		for (int i = 0, len = line.length(); i < len; i++) {
			int idx = INDENT_CHARS.indexOf(line.charAt(i));
			if (idx == -1)
				break;
			indent += INDENT_WIDTH[idx];
			if (indent == indentWidth)
				return line.substring(i + 1);
		}
		return line;
	}

	/**
	 * Reads the snippet and returns the lines in a List.
	 *
	 * @param snippetId The id of the snippet.
	 * @return A List of lines.
	 * @throws IOException if something goes wrong.
	 */
	private Map<Integer, List<String>> readLines(String snippetId) throws IOException {
		Map<Integer, List<String>> result = new LinkedHashMap<Integer, List<String>>();

		LineNumberReader lineReader = new LineNumberReader(reader);
		try {
			String line;
			List<String> lines = null;
			boolean hasSnippetId = !StringUtils.isEmpty(snippetId), capture = !hasSnippetId;

			while ((line = lineReader.readLine()) != null) {

				if (capture) {
					if (lines == null)
						result.put(lineReader.getLineNumber(), lines = new ArrayList<String>(64));

					if (hasSnippetId && isEnd(snippetId, line)) {
						lines = null;
						capture = false;
					} else
						lines.add(line);
				} else
					capture = hasSnippetId && isStart(snippetId, line);
			}
		} finally {
			lineReader.close();
		}

		return result;
	}

	/**
	 * Determines if the given line is a start demarcator.
	 *
	 * @param snippetId the id of the snippet.
	 * @param line	  the line.
	 * @return True, if the line is a start demarcator.
	 */

	protected boolean isStart(String snippetId, String line) {
		return isDemarcator(snippetId, "START", line);
	}

	/**
	 * Determines if the given line is a demarcator.
	 *
	 * @param snippetId the id of the snippet.
	 * @param what	  Identifier for the demarcator.
	 * @param line	  the line.
	 * @return True, if the line is a start demarcator.
	 */
	protected boolean isDemarcator(String snippetId, String what, String line) {
		String upper = line.toUpperCase(Locale.ENGLISH);
		return upper.contains(what.toUpperCase(Locale.ENGLISH))
				&& upper.contains("SNIPPET")
				&& line.contains(snippetId);
	}

	/**
	 * Determines if the given line is an end demarcator.
	 *
	 * @param snippetId the id of the snippet.
	 * @param line	  the line.
	 * @return True, if the line is an end demarcator.
	 */
	protected boolean isEnd(String snippetId, String line) {
		return isDemarcator(snippetId, "END", line);
	}
}
