/*
 * Copyright 2010 - org.tinyjee.maven
 *
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */

package org.tinyjee.maven.dim.sources;

import java.io.File;
import java.io.IOException;
import java.io.Reader;
import java.io.StringReader;
import java.net.URL;
import java.util.List;
import java.util.Map;

import static org.tinyjee.maven.dim.sources.URLSource.getSnippetId;
import static org.tinyjee.maven.dim.utils.ResolveUtils.resolveClass;

/**
 * Adapts a source class to the source interface.
 *
 * @author Juergen_Kellerer, 2010-09-04
 * @version 1.0
 */
public class SourceClassAdapter implements Source {

	/**
	 * Defines the map key to use for retrieving the source content from the
	 * Map defined with the wrapped class.
	 * <p/>
	 * Note: A source class can decide to return an instance of <code>Map&lt;Integer, List&lt;String&gt;&gt;</code>
	 * when queried for the key defined by this constant.
	 */
	public static final String CONTENT_MAP_KEY = "content";

	final Class<?> cls;
	final File basePath;
	final Map requestParams;
	Content content;

	/**
	 * Creates a new instance of SourceClassAdapter.
	 *
	 * @param basePath	  the base path containing the "target/classes" folder.
	 * @param requestParams the params of the macro request (are optionally passed to the object instance).
	 * @param className	 the name of the class to adapt.
	 */
	public SourceClassAdapter(File basePath, Map requestParams, String className) {
		this.basePath = basePath;
		this.requestParams = requestParams;
		cls = resolveClass(basePath, className);
	}

	Object instantiateClass(File basePath, Map requestParameter) {
		try {
			try {
				return cls.getConstructor(File.class, Map.class).newInstance(basePath, requestParameter);
			} catch (NoSuchMethodException ignored) {
				try {
					return cls.getConstructor(Map.class).newInstance(requestParameter);
				} catch (NoSuchMethodException ignored1) {
					try {
						return cls.getConstructor(File.class).newInstance(basePath);
					} catch (NoSuchMethodException ignored2) {
						return cls.newInstance();
					}
				}
			}
		} catch (Exception e) {
			throw new RuntimeException("Failed to create an instance of the configured sourceClass '" + cls + '\'', e);
		}
	}

	/**
	 * {@inheritDoc}
	 */
	public URL getURL() {
		return null;
	}

	/**
	 * {@inheritDoc}
	 */
	@SuppressWarnings("unchecked")
	public Content getContent() {
		if (content == null) {
			final Object instance = instantiateClass(basePath, requestParams);

			if (instance instanceof Content)
				content = (Content) instance;
			else if (instance instanceof Map)
				content = new MapContent((Map<String, Object>) instance);
			else if (instance instanceof Reader)
				content = new SimpleContent((Reader) instance);
			else
				content = new SimpleContent(new StringReader(instance.toString()));
		}
		return content;
	}

	private static class MapContent implements Content {

		private final Map<String, Object> source;

		MapContent(Map<String, Object> source) {
			this.source = source;
		}

		/**
		 * {@inheritDoc}
		 */
		public Map<String, Object> getParameters() {
			return source;
		}

		/**
		 * {@inheritDoc}
		 */
		@SuppressWarnings("unchecked")
		public Map<Integer, List<String>> getSections() {
			Object content = source.get(CONTENT_MAP_KEY);
			if (content instanceof Map)
				return (Map<Integer, List<String>>) content;
			else if (content != null)
				return ContentCodec.read(getSnippetId(source), content.toString());
			return null;
		}
	}

	private class SimpleContent implements Content {

		final Map<Integer, List<String>> sections;

		@SuppressWarnings("unchecked")
		SimpleContent(Reader reader) {
			try {
				sections = ContentCodec.read(getSnippetId(requestParams), reader);
			} catch (IOException e) {
				throw new RuntimeException(e);
			}
		}

		/**
		 * {@inheritDoc}
		 */
		@SuppressWarnings("unchecked")
		public Map<String, Object> getParameters() {
			return requestParams;
		}

		/**
		 * {@inheritDoc}
		 */
		public Map<Integer, List<String>> getSections() throws IOException {
			return sections;
		}
	}
}
