/*
 * Copyright 2010 - org.tinyjee.maven
 *
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */

package org.tinyjee.maven.dim.sources;

import org.apache.maven.doxia.logging.Log;
import org.apache.velocity.Template;
import org.apache.velocity.VelocityContext;
import org.apache.velocity.app.VelocityEngine;
import org.apache.velocity.context.Context;
import org.apache.velocity.runtime.RuntimeServices;
import org.apache.velocity.runtime.log.LogChute;
import org.apache.velocity.tools.ToolManager;
import org.apache.velocity.tools.config.EasyFactoryConfiguration;
import org.codehaus.plexus.util.FileUtils;
import org.tinyjee.maven.dim.utils.MacroLoggers;

import java.io.File;
import java.io.IOException;
import java.io.StringReader;
import java.io.StringWriter;
import java.net.URL;
import java.util.List;
import java.util.Map;

import static org.apache.velocity.runtime.RuntimeConstants.FILE_RESOURCE_LOADER_PATH;
import static org.apache.velocity.runtime.RuntimeConstants.RUNTIME_LOG_LOGSYSTEM;

/**
 * Wraps a source content that contains a velocity template and executes it on using the Source interfaces.
 *
 * @author Juergen_Kellerer, 2010-09-04
 * @version 1.0
 */
public class TemplateSource implements Source {

	static final String TEMPLATE_ENCODING = "UTF-8";

	private static final ThreadLocal<VelocityEngine> engineEnvironment = new ThreadLocal<VelocityEngine>() {
		@Override
		protected VelocityEngine initialValue() {
			try {
				final VelocityEngine ve = new VelocityEngine();
				File dummyFile = File.createTempFile("dummy", ".vm"),
						templateFolder = dummyFile.getParentFile();
				deleteTemporaryFile(dummyFile);

				ve.setProperty(FILE_RESOURCE_LOADER_PATH, templateFolder.getAbsolutePath());
				ve.setProperty(RUNTIME_LOG_LOGSYSTEM, new TemplateLogger());
				ve.init();

				return ve;
			} catch (Exception e) {
				throw new RuntimeException(e);
			}
		}
	};

	static final ToolManager VELOCITY_TOOL_MANAGER;
	static {
		EasyFactoryConfiguration config = new EasyFactoryConfiguration();
		config.toolbox("application").
				tool("org.apache.velocity.tools.generic.AlternatorTool").
				tool("org.apache.velocity.tools.generic.ClassTool").
				tool("org.apache.velocity.tools.generic.ComparisonDateTool").
				tool("org.apache.velocity.tools.generic.ConversionTool").
				tool("org.apache.velocity.tools.generic.DisplayTool").
				tool("org.apache.velocity.tools.generic.EscapeTool").
				tool("org.apache.velocity.tools.generic.FieldTool").
				tool("org.apache.velocity.tools.generic.MathTool").
				tool("org.apache.velocity.tools.generic.NumberTool").
				tool("org.apache.velocity.tools.generic.SortTool").
				tool("org.apache.velocity.tools.generic.ListTool").
				tool("org.apache.velocity.tools.generic.ResourceTool");

		config.autoLoad();

		ToolManager toolManager = new ToolManager(false, false);
		toolManager.configure(config);
		toolManager.setUserCanOverwriteTools(true);
		VELOCITY_TOOL_MANAGER = toolManager;
	}

	private static void deleteTemporaryFile(File file) {
		if (file != null && !file.delete()) {
			MacroLoggers.getLogger().warn("Failed to delete temporary file: " + file +
					", trying to delete on exit.");
			file.deleteOnExit();
		}
	}

	final Source source;
	final URL sourceURL;
	final Template template;
	final VelocityEngine engine = engineEnvironment.get();
	Content content;

	/**
	 * Constructs a new TemplateSource.
	 *
	 * @param sourceURL the original source URL (used for naming purposes).
	 * @param source	the source containing the template source and context params.
	 */
	public TemplateSource(URL sourceURL, Source source) {
		this.source = source;
		this.sourceURL = sourceURL;

		URLSource urlSource = source instanceof URLSource ? (URLSource) source : null;
		if (urlSource != null)
			urlSource.setIgnoreSnippetId(true);

		File templateFile = null;
		try {
			String tempNamePrefix = FileUtils.basename(new File(sourceURL.getFile()).getName());
			templateFile = File.createTempFile(tempNamePrefix, ".vm");

			FileUtils.fileWrite(templateFile.getAbsolutePath(), TEMPLATE_ENCODING,
					ContentCodec.write(source.getContent().getSections()));

			template = engine.getTemplate(templateFile.getName(), TEMPLATE_ENCODING);
		} catch (Exception e) {
			throw new RuntimeException(e);
		} finally {
			if (urlSource != null)
				urlSource.setIgnoreSnippetId(false);
			deleteTemporaryFile(templateFile);
		}
	}

	/**
	 * {@inheritDoc}
	 */
	public URL getURL() {
		return sourceURL;
	}

	/**
	 * {@inheritDoc}
	 */
	public Content getContent() {
		if (content == null) {
			final Content sourceContent = source.getContent();
			content = new Content() {
				public Map<String, Object> getParameters() {
					return sourceContent.getParameters();
				}

				public Map<Integer, List<String>> getSections() throws IOException {
					final Map<String, Object> parameters = getParameters();
					final Context context = new VelocityContext(parameters, VELOCITY_TOOL_MANAGER.createContext());

					StringWriter buffer = new StringWriter(2048);
					template.merge(context, buffer);

					Log log = MacroLoggers.getInstance().get();
					if (log.isDebugEnabled())
						log.debug("Result after Template Processing:\n" + buffer.toString());

					String snippetId = URLSource.getSnippetId(parameters);
					return ContentCodec.read(snippetId, new StringReader(buffer.toString()));
				}
			};
		}
		return content;
	}

	/**
	 * Translates velocity logs to Maven/Doxia logs.
	 */
	private static class TemplateLogger implements LogChute {

		/**
		 * {@inheritDoc}
		 */
		public void init(RuntimeServices runtimeServices) throws Exception {
		}

		/**
		 * {@inheritDoc}
		 */
		public void log(int i, String s) {
			log(i, s, null);
		}

		/**
		 * {@inheritDoc}
		 */
		public void log(int i, String s, Throwable throwable) {
			Log log = MacroLoggers.getInstance().get();
			switch (i) {
				case TRACE_ID:
				case DEBUG_ID:
					if (throwable == null) log.debug(s);
					else log.debug(s, throwable);
					break;
				case INFO_ID:
					if (throwable == null) log.info(s);
					else log.info(s, throwable);
					break;
				case WARN_ID:
					if (throwable == null) log.warn(s);
					else log.warn(s, throwable);
					break;
				case ERROR_ID:
					if (throwable == null) log.error(s);
					else log.error(s, throwable);
					break;
			}
		}

		/**
		 * {@inheritDoc}
		 */
		public boolean isLevelEnabled(int i) {
			Log log = MacroLoggers.getInstance().get();
			switch (i) {
				case TRACE_ID:
				case DEBUG_ID:
					return log.isDebugEnabled();
				case INFO_ID:
					return log.isInfoEnabled();
			}
			return true;
		}
	}
}
