/*
 * Copyright 2010 - org.tinyjee.maven
 *
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */

package org.tinyjee.maven.dim.sources;

import org.codehaus.plexus.util.IOUtil;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.lang.ref.SoftReference;
import java.net.URL;
import java.nio.charset.Charset;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import static org.tinyjee.maven.dim.IncludeMacroSignature.*;

/**
 * Wraps the logic of translating a source URL to the Source interface,
 * including in-memory caching of loaded source content (important if URL points to a public location).
 *
 * @author Juergen_Kellerer, 2010-09-06
 * @version 1.0
 */
public class URLSource implements Source {

	/**
	 * Specified the default charset to use when loading text content.
	 * (use the 'charset' parameter to override)
	 */
	public static final Charset DEFAULT_ENCODING = Charset.forName("UTF-8");

	/**
	 * Defines the max size of the URL content cache.
	 */
	public static int maxCacheSize = 1024;

	/**
	 * Defines a static URL content cache using soft references (to avoid OutOfMemory errors).
	 */
	public static final Map<String, SoftReference<byte[]>> CACHE = new HashMap<String, SoftReference<byte[]>>();

	/**
	 * Extracts the snippet ID out of the given request params.
	 *
	 * @param params the request params of the macro request.
	 * @return the snippet ID if contained in the params, 'null' otherwise.
	 */
	public static String getSnippetId(Map<String, Object> params) {
		Object snippetId = params.get(PARAM_SNIPPET_ID);
		return snippetId == null ? null : snippetId.toString();
	}

	/**
	 * Returns the content stream of the given source URL.
	 *
	 * @param sourceURL the source URL to open.
	 * @param noCache   whether the content shall be cached / retrieved from the cache or not.
	 * @return the content stream of the given source URL.
	 * @throws IOException in case of the IO transfer failed.
	 */
	public static synchronized InputStream getSource(URL sourceURL, boolean noCache) throws IOException {
		if (noCache)
			return sourceURL.openStream();

		final String urlString = sourceURL.toExternalForm();
		SoftReference<byte[]> ref = CACHE.get(urlString);
		byte[] content = ref == null ? null : ref.get();

		if (content == null) {
			content = IOUtil.toByteArray(sourceURL.openStream(), 128 * 1024);

			if (CACHE.size() > maxCacheSize) {
				for (Iterator<SoftReference<byte[]>> iterator = CACHE.values().iterator(); iterator.hasNext(); ) {
					if (iterator.next().get() == null)
						iterator.remove();
				}
			}

			if (CACHE.size() < maxCacheSize)
				CACHE.put(urlString, new SoftReference<byte[]>(content));
		}

		return new ByteArrayInputStream(content);
	}

	final URL source;
	final Map<String, Object> params;

	boolean ignoreSnippetId;

	final Content content = new Content() {
		/**
		 * {@inheritDoc}
		 */
		public Map<String, Object> getParameters() {
			return params;
		}

		/**
		 * {@inheritDoc}
		 */
		public Map<Integer, List<String>> getSections() throws IOException {
			Object charsetName = params.get(PARAM_CHARSET);
			boolean noCache = Boolean.valueOf(String.valueOf(params.get(PARAM_NO_CACHE)));
			Charset charset = charsetName == null ? DEFAULT_ENCODING : Charset.forName(charsetName.toString());

			String snippetId = ignoreSnippetId ? null : getSnippetId(params);

			InputStreamReader streamReader = new InputStreamReader(getSource(source, noCache), charset);
			try {
				return ContentCodec.read(snippetId, streamReader);
			} finally {
				streamReader.close();
			}
		}
	};

	/**
	 * Constructs a new URLSource for the given source URL and request params.
	 *
	 * @param source the source URL pointing to the source content.
	 * @param params the params of the macro request.
	 */
	public URLSource(URL source, Map<String, Object> params) {
		this.source = source;
		this.params = params;
	}

	public boolean isIgnoreSnippetId() {
		return ignoreSnippetId;
	}

	public void setIgnoreSnippetId(boolean ignoreSnippetId) {
		this.ignoreSnippetId = ignoreSnippetId;
	}

	/**
	 * {@inheritDoc}
	 */
	public URL getURL() {
		return source;
	}

	/**
	 * {@inheritDoc}
	 */
	public Content getContent() {
		return content;
	}
}
