/*
 * Copyright 2010 - org.tinyjee.maven
 *
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */

package org.tinyjee.maven.dim.utils;

import com.thoughtworks.qdox.JavaDocBuilder;
import com.thoughtworks.qdox.model.*;

import java.io.File;
import java.io.IOException;
import java.util.*;

import static org.tinyjee.maven.dim.utils.ResolveUtils.findSource;

/**
 * InterfaceScanner is a 'source-class' compatible helper class that scans an interface class
 * (at source level including JavaDoc) and makes the details available to the source parameters
 * map.
 * <p/>
 * <b>Usage:</b><div><code>
 * %{include|source=template.vm|source-class=org.tinyjee.maven.dim.utils.InterfaceScanner|interface=net.sf.MyClass}
 * </code></div>
 * or when using the alias:<div><code>
 * %{include|source=template.vm|source-interface-api-scan=net.sf.MyClass}
 * </code></div>
 * <p/>
 * Note: Annotations that are applied to fields, properties or methods are pre-processed for simplified access
 * within a velocity template and are stored within list rows using a map. Annotation map keys are the
 * simple name of the mapped annotation starting with the &#64; and the values are maps of annotation
 * parameter name to value pairs.
 *
 * @author Juergen_Kellerer, 2010-09-08
 * @version 1.0
 */
public class InterfaceScanner extends HashMap<String, Object> {

	private static final long serialVersionUID = -4069212921446859127L;

	/**
	 * Implements the "{@link InterfaceScanner#PARAM_ALIAS}" alias functionality.
	 */
	public static class AliasHandler extends AbstractAliasHandler {
		/**
		 * Constructs the handler (Note: Is called by
		 * {@link org.tinyjee.maven.dim.RequestParameterTransformer#TRANSFORMERS}).
		 */
		public AliasHandler() {
			super(PARAM_ALIAS, PARAM_INTERFACE, InterfaceScanner.class.getName());
		}
	}

	/**
	 * Defines a shortcut for the request properties 'source-class' and 'interface'.
	 * <p/>
	 * If this parameter is present inside the request parameters list, the system
	 * will effectively behave as if 'source-class' and 'interface' where set
	 * separately.
	 */
	public static final String PARAM_ALIAS = "source-interface-api-scan";

	/**
	 * Is the expected input parameter that points to the class or interface to scan.
	 * <p/>
	 * The class can be specified as file (with .java as extension) or as class name.
	 * Inner classes are not supported.
	 */
	public static final String PARAM_INTERFACE = "interface";

	/**
	 * Is filled with a the JavaDocComment of the specified class.
	 */
	public static final String OUT_PARAM_COMMENT = "comment";

	/**
	 * Is filled with a the qdox..JavaClass instance of the specified class.
	 */
	public static final String OUT_PARAM_TYPE = "type";

	/**
	 * Is filled with a list of public or protected class fields.
	 * Every list row is a map that follows the format:<br/>
	 * <code>("type":type, "name":fieldName, "value":initializationExpression, "comment":JavaDocComment,
	 * "field":qdox..JavaField, "annotations":Map&lt;String,Map&lt;String,String&gt;&gt;)</code>
	 */
	public static final String OUT_PARAM_FIELDS = "fields";

	/**
	 * Is filled with a list of all class fields that are not static and final.
	 * Every list row is a map that follows the format:<br/>
	 * <code>("type":type, "name":fieldName, "value":initializationExpression, "comment":JavaDocComment,
	 * "field":qdox..JavaField, "annotations":Map&lt;String,Map&lt;String,String&gt;&gt;)</code>
	 */
	public static final String OUT_PARAM_DECLARED_FIELDS = "declaredFields";

	/**
	 * Is filled with a list of public or protected class constants (static final fields).
	 * Every list row is a map that follows the format:<br/>
	 * <code>("type":type, "name":fieldName, "value":initializationExpression, "comment":JavaDocComment,
	 * "field":qdox..JavaField, "annotations":Map&lt;String,Map&lt;String,String&gt;&gt;)</code>
	 */
	public static final String OUT_PARAM_CONSTANTS = "constants";

	/**
	 * Is filled with a list of all class constants (static final fields).
	 * Every list row is a map that follows the format:<br/>
	 * <code>("type":type, "name":fieldName, "value":initializationExpression, "comment":JavaDocComment,
	 * "field":qdox..JavaField, "annotations":Map&lt;String,Map&lt;String,String&gt;&gt;)</code>
	 */
	public static final String OUT_PARAM_DECLARED_CONSTANTS = "declaredConstants";

	/**
	 * Is filled with a list of bean properties.
	 * Every list row is a map that follows the format:<br/>
	 * <code>("type":propertyType, "name":propertyName, "comment":GetterJavaDocComment,
	 * "setterComment":SetterJavaDocComment, "property":qdox..BeanProperty,
	 * "getter":qdox..JavaMethod, "setter":qdox..JavaMethod, "field":qdox..JavaField,
	 * "value":fieldInitializationExpression, "annotations":Map&lt;String,Map&lt;String,String&gt;&gt;)</code>
	 * <p/>
	 * Note: The annotations map contains a combined map of all properties that were either specified on getter,
	 * setter or field.
	 */
	public static final String OUT_PARAM_PROPERTIES = "properties";

	/**
	 * Is filled with a list of abstract or interface methods.
	 * Every list row is a map that follows the format:<br/>
	 * <code>("type":returnType, "name":methodName, "signature":callSignature, "comment":JavaDocComment,
	 * "method":qdox..JavaMethod, "annotations":Map&lt;String,Map&lt;String,String&gt;&gt;,
	 * "parameterAnnotations":Map&lt;String,Map&lt;String,Map&lt;String,String&gt;&gt;&gt;)</code>
	 */
	public static final String OUT_PARAM_INTERFACE_METHODS = "interfaceMethods";

	/**
	 * Is filled with a list of public or protected methods.
	 * Every list row is a map that follows the format:<br/>
	 * <code>("type":returnType, "name":methodName, "signature":callSignature, "comment":JavaDocComment,
	 * "method":qdox..JavaMethod, "annotations":Map&lt;String,Map&lt;String,String&gt;&gt;,
	 * "parameterAnnotations":Map&lt;String,Map&lt;String,Map&lt;String,String&gt;&gt;&gt;)</code>
	 */
	public static final String OUT_PARAM_METHODS = "methods";

	/**
	 * Constructs a new interface scanner.
	 *
	 * @param baseDir	   the base dir to of the maven module.
	 * @param requestParams the request params of the macro call.
	 */
	public InterfaceScanner(File baseDir, Map<String, Object> requestParams) {
		this(baseDir, (String) requestParams.get(PARAM_INTERFACE));
		putAll(requestParams);
	}

	/**
	 * Constructs a new interface scanner for the given interface or abstract class.
	 *
	 * @param baseDir   the base dir to of the maven module.
	 * @param className the class describing the interface.
	 */
	public InterfaceScanner(File baseDir, String className) {
		super();
		try {
			init(getSource(baseDir, className));
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
	}

	/**
	 * Initializes the map values.
	 *
	 * @param source the source to scan.
	 */
	protected void init(final JavaSource source) {
		final JavaClass javaClass = source.getClasses()[0];

		scanFieldsAndConstants(javaClass);
		scanProperties(javaClass);
		scanMethods(javaClass);

		put(OUT_PARAM_TYPE, javaClass);
		put(OUT_PARAM_COMMENT, javaClass.getComment());
	}

	/**
	 * Loads the given class from the module's source directory.
	 *
	 * @param baseDir   the module's base dir (containing the 'src/main/java' folder).
	 * @param className the name of the class.
	 * @return The parsed JavaSource instance.
	 * @throws IOException in case of the java source cannot be found or read.
	 */
	protected JavaSource getSource(File baseDir, String className) throws IOException {
		JavaDocBuilder builder = new JavaDocBuilder();
		className = className.replace('.', '/');
		if (className.contains("$"))
			className = className.substring(0, className.indexOf('$'));
		if (!className.endsWith(".java"))
			className += ".java";
		return builder.addSource(findSource(baseDir, className));
	}

	/**
	 * Scans all relevant methods of the specified class and sets the results as map values.
	 *
	 * @param javaClass the java class to scan.
	 */
	protected void scanMethods(JavaClass javaClass) {
		final List<Map<?, ?>> interfaceMethods = new ArrayList<Map<?, ?>>();
		put(OUT_PARAM_INTERFACE_METHODS, interfaceMethods);
		final List<Map<?, ?>> methods = new ArrayList<Map<?, ?>>();
		put(OUT_PARAM_METHODS, methods);

		for (JavaMethod method :javaClass.getMethods()) {
			boolean nonVisibleMethod = !method.isPublic() && !method.isProtected();
			if (nonVisibleMethod || method.isConstructor() || method.isPropertyAccessor() || method.isPropertyMutator())
				continue;

			Map<String, Map> parameterAnnotations = new LinkedHashMap<String, Map>();
			if (method.getParameters() != null) {
				for (JavaParameter parameter : method.getParameters())
					parameterAnnotations.put(parameter.getName(), asMap(parameter.getAnnotations()));
			}

			Map<?, ?> row = asMap(
					"type", typeToString(method.getReturns()),
					"name", method.getName(),
					"signature", method.getCallSignature(),
					"comment", method.getComment(),
					"method", method,
					"annotations", asMap(method.getAnnotations()),
					"parameterAnnotations", parameterAnnotations);

			if (method.isAbstract())
				interfaceMethods.add(row);
			else
				methods.add(row);
		}
	}

	/**
	 * Scans all relevant bean properties of the specified class and sets the results as map values.
	 *
	 * @param javaClass the java class to scan.
	 */
	protected void scanProperties(JavaClass javaClass) {
		final List<Map<?, ?>> properties = new ArrayList<Map<?, ?>>();
		put(OUT_PARAM_PROPERTIES, properties);

		for (BeanProperty property : javaClass.getBeanProperties()) {
			final JavaMethod getter = property.getAccessor();
			final JavaMethod setter = property.getMutator();
			final JavaField field = javaClass.getFieldByName(property.getName());

			final Map<String, Map> annotations = new LinkedHashMap<String, Map>();
			if (getter != null)
				annotations.putAll(asMap(getter.getAnnotations()));
			if (setter != null)
				annotations.putAll(asMap(setter.getAnnotations()));
			if (field != null)
				annotations.putAll(asMap(field.getAnnotations()));

			final Map<?, ?> row = asMap(
					"type", typeToString(property.getType()),
					"name", property.getName(),
					"comment", getter == null ? "" : getter.getComment(),
					"setterComment", setter == null ? "" : setter.getComment(),
					"property", property,
					"getter", getter,
					"setter", setter,
					"field", field,
					"value", field == null ? "" : field.getInitializationExpression(),
					"annotations", annotations);

			properties.add(row);
		}
	}

	@Override
	public Object clone() {
		return super.clone();
	}

	/**
	 * Scans all relevant fields and constants of the specified class and sets the results as map values.
	 *
	 * @param javaClass the java class to scan.
	 */
	protected void scanFieldsAndConstants(JavaClass javaClass) {
		final List<Map<?, ?>> constants = new ArrayList<Map<?, ?>>(), declaredConstants = new ArrayList<Map<?, ?>>();
		put(OUT_PARAM_CONSTANTS, constants);
		put(OUT_PARAM_DECLARED_CONSTANTS, declaredConstants);
		final List<Map<?, ?>> fields = new ArrayList<Map<?, ?>>(), declaredFields = new ArrayList<Map<?, ?>>();
		put(OUT_PARAM_FIELDS, fields);
		put(OUT_PARAM_DECLARED_FIELDS, declaredFields);

		boolean classIsInterface = javaClass.isInterface();

		for (JavaField field : javaClass.getFields()) {
			boolean fieldIsConstant = classIsInterface || (field.isFinal() && field.isStatic());

			Map<?, ?> row = asMap(
					"type", typeToString(field.getType()),
					"name", field.getName(),
					"value", field.getInitializationExpression(),
					"comment", field.getComment(),
					"field", field,
					"annotations", asMap(field.getAnnotations()));

			(fieldIsConstant ? declaredConstants : declaredFields).add(row);

			if (!field.isPublic() && !field.isProtected() && !classIsInterface)
				continue;

			(fieldIsConstant ? constants : fields).add(row);
		}
	}

	/**
	 * Converts the given type to string.
	 *
	 * @param type the type to convert.
	 * @return the string representation or an empty string.
	 */
	protected static String typeToString(Type type) {
		return type == null ? "" : type.toGenericString();
	}

	/**
	 * Converts the given set of annotations to a map, mapping the simple name of the annotation
	 * against a key=>value map of the specified parameters (the simple name starts with '@').
	 *
	 * @param annotations the annotations to convert to a map.
	 * @return a map mapping the simple name of the annotation to its parameters.
	 */
	protected static Map<String, Map> asMap(Annotation... annotations) {
		Map<String, Map> result = new LinkedHashMap<String, Map>(annotations.length);
		for (Annotation annotation : annotations) {
			String name = annotation.getType().getFullyQualifiedName();
			name = "@".concat(name.substring(name.lastIndexOf('.') + 1));

			Map<Object, Object> map = new LinkedHashMap(annotation.getNamedParameterMap());
			for (Map.Entry<Object, Object> entry : map.entrySet())
				if (entry.getValue() != null)
					entry.setValue(entry.getValue().toString().trim());

			result.put(name, map);
		}
		return result;
	}

	/**
	 * Converts the given key and value pairs to a Map that is linked by key and index (=> associative array).
	 *
	 * @param keyValuePairs the key and value pairs to convert.
	 * @return a Map that is linked by key and index.
	 */
	protected static Map<Object, Object> asMap(Object... keyValuePairs) {
		if (keyValuePairs.length % 2 != 0)
			throw new IllegalArgumentException("The key and value input array length must be a multiple of 2.");

		final Map<Object, Object> map = new LinkedHashMap<Object, Object>(keyValuePairs.length),
				assocMap = new LinkedHashMap<Object, Object>(keyValuePairs.length / 2);

		for (int i = 0; i < keyValuePairs.length; i += 2) {
			Object key = String.valueOf(keyValuePairs[i]), value = keyValuePairs[i + 1];
			map.put(i / 2, value);
			assocMap.put(key, value);
		}

		map.putAll(assocMap);

		return map;
	}
}
