/*
 * Copyright 2010 - org.tinyjee.maven
 *
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */

package org.tinyjee.maven.dim.utils;

import org.codehaus.plexus.util.FileUtils;

import java.io.File;
import java.io.IOException;
import java.net.URL;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;

import static org.tinyjee.maven.dim.utils.ResolveUtils.findSource;

/**
 * PropertiesLoader is a 'source-class' compatible helper class that can load a properties file
 * and add or overwrite the request properties by the properties contained in the file.
 *
 * <p/>
 * <b>Usage:</b><div><code>
 * %{include|source=template.vm|source-class=org.tinyjee.maven.dim.utils.PropertiesLoader|properties=props.(xml|properties)}
 * </code></div>
 * or when using the alias:<div><code>
 * %{include|source=template.vm|source-properties=props.(xml|properties)}
 * </code></div>
 *
 * @author Juergen_Kellerer, 2010-09-08
 * @version 1.0
 */
public class PropertiesLoader extends HashMap<String, Object> {

	private static final long serialVersionUID = 7451224247313497738L;

	/**
	 * Implements the "{@link PropertiesLoader#PARAM_ALIAS}" alias functionality.
	 */
	public static class AliasHandler extends AbstractAliasHandler {
		/**
		 * Constructs the handler (Note: Is called by
		 * {@link org.tinyjee.maven.dim.RequestParameterTransformer#TRANSFORMERS}).
		 */
		public AliasHandler() {
			super(PARAM_ALIAS, PARAM_PROPERTIES, PropertiesLoader.class.getName());
		}
	}

	/**
	 * Defines a shortcut for the request properties 'source-class' and 'properties'.
	 * <p/>
	 * If this parameter is present inside the request parameters list, the system
	 * will effectively behave as if 'source-class' and 'properties' where set
	 * separately.
	 */
	public static final String PARAM_ALIAS = "source-properties";

	/**
	 * Sets the path to the properties file to load inside the request parameters.
	 * <p/>
	 * The specified file is retrieved in exactly the same way as if the "<code>source</code>"
	 * parameter would have been used (see Macro reference).
	 * <p/>
	 * Note: Properties can use the traditional <code>.properties</code> format or
	 * they can use the <code>.xml</code> format. The format is selected by file extension.
	 */
	public static final String PARAM_PROPERTIES = "properties";

	/**
	 * Constructs a new properties loader searching for properties using the given base path.
	 *
	 * @param baseDir	   the base dir to of the maven module.
	 * @param requestParams the request params of the macro call.
	 */
	@SuppressWarnings("unchecked")
	public PropertiesLoader(File baseDir, Map<String, Object> requestParams) {
		super();
		putAll(requestParams);

		URL propertiesURL = findSource(baseDir, (String) requestParams.get(PARAM_PROPERTIES));

		try {
			Properties properties = new Properties();
			if ("xml".equalsIgnoreCase(FileUtils.extension(propertiesURL.getFile())))
				properties.loadFromXML(propertiesURL.openStream());
			else
				properties.load(propertiesURL.openStream());

			for (Map.Entry<Object, Object> entry : properties.entrySet())
				put((String) entry.getKey(), entry.getValue());
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
	}

	@Override
	public Object clone() {
		return super.clone();
	}
}
